# SPDX-FileCopyrightText: 2022 Alec Delaney
#
# SPDX-License-Identifier: MIT

"""
The main script handling CLI interactions for ``circlink``.

Author(s): Alec Delaney (Tekktrik)
"""

import os
import shutil
import sys

from circup import find_device
from tabulate import tabulate
from typer import Argument, Exit, Option, Typer

from circlink import (
    APP_DIRECTORY,
    __version__,
    ensure_app_folder_setup,
    get_settings,
)
from circlink.backend import clear_backend, start_backend, stop_backend
from circlink.cli import config, workspace
from circlink.ledger import iter_ledger_entries
from circlink.link import CircuitPythonLink, get_links_header, get_links_list

# Prevent running on non-POSIX systems that don't have os.fork()
if os.name != "posix":
    print("circlink is currently only available for Linux and macOS")
    sys.exit(1)

# Create the Typer apps
app = Typer(
    add_completion=False,
    no_args_is_help=True,
    help="Autosave local files to your CircuitPython board",
)
app.add_typer(config.config_app, name="config")
app.add_typer(workspace.workspace_app, name="workspace")


@app.command()
def start(
    read_path: str = Argument(..., help="The read path/pattern of file(s) to save"),
    write_path: str = Argument(
        ...,
        help="The write path of the directory to write files, relative to the CircuitPython board",
    ),
    *,
    path: bool = Option(
        False,
        "--path",
        "-p",
        help="Designate the write path as absolute or relative to the current directory",
    ),
    name: str = Option("", "--name", "-n", help="A name for the new link"),
    recursive: bool = Option(
        False, "--recursive", "-r", help="Whether the link glob pattern is recursive"
    ),
    wipe_dest: bool = Option(
        False,
        "--wipe-dest",
        "-w",
        help="Wipe the write destination recursively before starting the link",
    ),
    skip_presave: bool = Option(
        False,
        "--skip-presave",
        "-s",
        help="Skip the inital save and write performed when opening a link",
    ),
) -> None:
    """Start a CircuitPython link."""
    start_backend(
        read_path,
        write_path,
        os.getcwd(),
        path=path,
        name=name,
        recursive=recursive,
        wipe_dest=wipe_dest,
        skip_presave=skip_presave,
    )
    workspace.set_cws_name("")


@app.command()
def stop(
    link_id: str = Argument(..., help="Link ID / 'last' / 'all'"),
    clear_flag: bool = Option(
        False,
        "--clear",
        "-c",
        help="Clear the history of the specified link(s) as well",
    ),
) -> bool:
    """Stop a CircuitPython link."""
    # If stopping all links, stop links using the "last" option until done
    if link_id == "all":
        link_entries = get_links_list("*")
        for link_entry in link_entries:
            stop_backend(link_entry[0], hard_fault=False)
            if clear_flag:
                clear_backend(link_entry[0], hard_fault=False)
        raise Exit()

    # If stopping the last link, calculate its link ID
    if link_id == "last":
        link_id = str(CircuitPythonLink.get_next_link_id() - 1)
        if link_id == "0":
            print("There are no links in the history")
            raise Exit(1)

    # Detect if the link ID is not valid
    try:
        link_id = int(link_id)
    except ValueError as err:
        print('Link ID must be the ID, "last", or "all"')
        raise Exit(1) from err

    # Stop the link, clear as well if requested
    stop_backend(link_id)
    if clear_flag:
        clear_backend(link_id)


@app.command()
def clear(
    link_id: str = Argument(..., help="Link ID / 'last' / 'all'"),
    *,
    force: bool = Option(
        False, "--force", "-f", help="Ignore warning and force clear from history"
    ),
) -> None:
    """Clear the link from the history."""
    # If clearing all links, repetitively clear the last link
    if link_id == "all":
        link_entries = get_links_list("*")
        for link_entry in link_entries:
            clear_backend(link_entry[0], force=force, hard_fault=False)
        raise Exit()

    # If clearing the last link link, calculate its link ID
    if link_id == "last":
        link_id = str(CircuitPythonLink.get_next_link_id() - 1)
        if link_id == "0":
            return

    # Detect if the link ID is not valid
    try:
        link_id = int(link_id)
    except ValueError as err:
        print('Link ID must be the ID, "last", or "all"')
        raise Exit(1) from err

    # Clear the link
    clear_backend(link_id, force=force)


@app.command()
def view(
    link_id: str = Argument("all", help="Link ID / 'last' / 'all' (default)"),
    *,
    abs_paths: bool = Option(
        False, "--abs-path", "-a", help="Show the read path as absolute"
    ),
) -> None:
    """List links in the history."""
    # For recursion purposes, note whether link ID is "last"
    last_requested_flag = False

    # Handle cases of link ID being "all" or "last", or just parse link ID
    if link_id == "all":
        pattern = "*"
    elif link_id == "last":
        last_requested_flag = True
        link_id = str(CircuitPythonLink.get_next_link_id() - 1)
        pattern = "link" + link_id + ".json"
        if link_id == "0":
            pattern = "*"
    else:
        try:
            int(link_id)
            pattern = "link" + link_id + ".json"
        except ValueError:
            print('Please use a valid link ID, "last", or "all" (default)')

    # Discard the link base directory for printing purposes
    link_infos = [x[:-1] for x in get_links_list(pattern, abs_paths=abs_paths)]

    # Handle if no links available
    if not link_infos:
        if link_id == "all" or last_requested_flag:
            print("No links in the history to view")
            raise Exit()
        print("This link ID is not in the history")
        raise Exit(1)

    # Prepare process ID depending on config settings
    show_list = get_links_header()
    if not get_settings()["display"]["info"]["process-id"]:
        show_list = list(show_list)
        show_list.remove("Process ID")
        link_infos = [entry[:-1] for entry in link_infos]

    # Print the table with the format based on config settings
    print(
        tabulate(
            link_infos,
            headers=show_list,
            tablefmt=get_settings()["display"]["table"]["format"],
        )
    )


@app.command()
def restart(link_id: str = Argument(..., help="Link ID / 'last' / 'all'")) -> None:
    """Restart a link."""
    # Handle cases of "all" or "last", or parse link ID
    if link_id == "all":
        pattern = "*"
    elif link_id == "last":
        link_id = str(CircuitPythonLink.get_next_link_id() - 1)
        pattern = "link" + link_id + ".json"
        if link_id == "0":
            pattern = "*"
    else:
        try:
            int(link_id)
            pattern = "link" + link_id + ".json"
        except ValueError as err:
            print('Please use a valid link ID, "last", or "all"')
            raise Exit(1) from err

    # Get the list of links in history if possible
    link_list = get_links_list(pattern)
    if not link_list:
        print("There are no links in the history to restart")
        raise Exit(1)

    # Attempt to restart and clear the link if it's not active
    for link in link_list:
        if link[2]:
            print(f"Link #{link[0]} is active, not restarting this link.")
        else:
            start_backend(
                str(link[3]),
                str(link[4]),
                link[-1],
                name=link[1],
                recursive=link[5],
                path=True,
            )
            clear_backend(link[0])


@app.command()
def detect() -> None:
    """Attempt to detect a CircuitPython board."""
    device = find_device()
    if device:
        print("CircuitPython device detected:", device)
    else:
        print("No CircuitPython device detected")


def about_cb() -> None:
    """Display information about circlink."""
    print("Originally built with love by Tekktrik")
    print("Happy hackin'!")
    raise Exit()


def version_cb() -> None:
    """Display the current version of circlink."""
    print(__version__)
    raise Exit()


@app.callback(invoke_without_command=True)
def callback(
    version: bool = Option(
        False, "--version", "-v", help="Display the version of circlink"
    ),
    about: bool = Option(False, "--about", "-a", help="A bit about circlink"),
    reset: bool = Option(
        False, "--reset", help="Reset the circlink configuration settings"
    ),
) -> None:
    """Display the current version of circlink."""
    ensure_app_folder_setup()

    if version:
        version_cb()
    if about:
        about_cb()
    if reset:
        reset_cb()


def reset_cb() -> None:
    """
    Reset the app directory.

    Useful if you upgrade circlink and there are breaking changes.
    """
    shutil.rmtree(APP_DIRECTORY)
    print("Removed circlink app directory, settngs and history deleted!")
    print("These will be created on next use of circlink.")
    print("Please check the integrity of any files handled by circlink.")
    raise Exit()


@app.command()
def ledger() -> None:
    """View the ledger of files controlled by links."""
    # Get the list of ledger entries if possible
    ledger_entries = list(iter_ledger_entries())
    if not ledger_entries:
        print("No files being tracked by circlink")
        raise Exit()

    # Display the process ID of links depending on settings
    table_headers = ("Write Path", "Link")
    if get_settings()["display"]["info"]["process-id"]:
        table_headers = table_headers + ("Process ID",)
    else:
        ledger_entries = [entry[:-1] for entry in ledger_entries]

    # Print the table with the format specified in config settings
    print(
        tabulate(
            ledger_entries,
            headers=table_headers,
            tablefmt=get_settings()["display"]["table"]["format"],
        )
    )



# SPDX-FileCopyrightText: 2022 Alec Delaney
#
# SPDX-License-Identifier: MIT

"""
Backend operations for circlink.

Author(s): Alec Delaney (Tekktrik)
"""

import os
import signal
import time
from datetime import datetime, timedelta

import psutil
from circup import find_device
from typer import Exit

from circlink.cli import workspace
from circlink.ledger import iter_ledger_entries, remove_from_ledger
from circlink.link import CircuitPythonLink, LedgerEntry


# pylint: disable=too-many-locals,too-many-branches
def start_backend(
    read_path: str,
    write_path: str,
    base_dir: str,
    *,
    path: bool = False,
    name: str = "",
    recursive: bool = False,
    wipe_dest: bool = False,
    skip_presave: bool = False,
) -> None:
    """Start a link (backend)."""
    # Only allow recursive setting with glob patterns
    if "*" not in read_path and recursive:
        print("--recursive can only be used with glob patterns!")
        raise Exit(1)

    # Attempt to find the CircuitPython board unless explicitly set otherwise
    if not path:
        device_path = find_device()
        if not device_path:
            print("Cound not auto-detect board path!")
            raise Exit(1)
        write_path = os.path.join(device_path, write_path)

    # Warn if the board (or write path in general) is not write accessible
    if not os.access(write_path, os.W_OK):
        print("Cannot write to the device or specified path")
        print("If using CircuitPython board, please ensure it is mounted")
        raise Exit(1)

    # Set the base directory is current directory if not provided
    base_dir = os.getcwd() if not base_dir else base_dir

    # Create the link object
    link = CircuitPythonLink(
        read_path,
        write_path,
        base_dir,
        name=name,
        recursive=recursive,
        wipe_dest=wipe_dest,
        skip_presave=skip_presave,
    )

    # Get the link ID and save the link
    link_id = link.link_id
    link.save_link()

    # Detect whether bad read path/pattern provided
    try:
        link.read_path.resolve().relative_to(base_dir)
    except ValueError as err:
        print(
            "Error occurred, please ensure the read file/pattern "
            "is relative to the current directory"
        )
        raise Exit(1) from err

    # Fork the process to begin start the link
    pid = os.fork()  # pylint: disable=no-member

    if pid:  # Current process, pid is that of spawned process

        # Save the link with the process ID
        link.process_id = pid
        link.save_link()

        # Attempt to wait for spawned process to be confirmed
        start_time = datetime.now()
        error_time = start_time + timedelta(seconds=5)
        while not link.confirmed:
            link = CircuitPythonLink.load_link_by_num(link_id)
            time.sleep(0.5)  # Slight delay
            if datetime.now() >= error_time:
                try:
                    os.kill(pid, signal.SIGTERM)
                except ProcessLookupError:
                    pass
                raise OSError("Could not start link process!")

        print(f"Started link #{link.link_id}")

    else:  # Spawned process, PID is 0

        # Wait for the process ID to be avaiable
        while not link.process_id:
            link = CircuitPythonLink.load_link_by_num(link_id)
            time.sleep(0.3)

        # Mark the link is confirmed and save it
        link.confirmed = True
        link.save_link()

        # Begin monitoring files, exiting successfully when completed or error if needed
        try:
            link.begin_monitoring()
        except FileNotFoundError:
            for file in link.get_files_monitored():
                ledger_entry = LedgerEntry(
                    str(file.resolve()), link.link_id, link.process_id
                )
                remove_from_ledger(ledger_entry)
            Exit(1)
        raise Exit()


def stop_backend(link_id: int, *, hard_fault: bool = True) -> bool:
    """Stop a link (backend)."""
    # Attempt to get the link by ID
    try:
        link = CircuitPythonLink.load_link_by_num(link_id)
    except FileNotFoundError as err:
        print(f"Link #{link_id} does not exist!")
        raise Exit(1) from err

    # Announce if the link is already stopped
    if link.stopped:
        print(f"Link #{link.link_id} is already stopped")
        if hard_fault:
            raise Exit()
        return False

    # Detect whether the link exists and is circlink
    try:
        circlink_process = psutil.Process(link.process_id)
        maybe_process = circlink_process.name() == "circlink"
    except psutil.NoSuchProcess:
        maybe_process = False

    if not maybe_process:
        print(
            f"Problem encountered stopping link #{link_id}!\n"
            "Asscoiated proess either does not exist, was already "
            "stopped, or isn't circlink.\n"
            "Consider using the clear command with the --force flag to "
            "clear it from the history."
        )
        raise Exit(1)

    # Mark the link for termination
    link.end_flag = True
    link.save_link()

    # Wait for confirmation that the link has stopped
    start_time = datetime.now()
    error_time = start_time + timedelta(seconds=5)
    while not link.stopped:
        link = CircuitPythonLink.load_link_by_num(link_id)
        time.sleep(0.1)  # Slight delay
        if datetime.now() >= error_time:
            print(f"Link #{link.link_id} could not be stopped!")
            if hard_fault:
                raise Exit(1)
            return False

    # Announce the link has stopped
    print(f"Stopped link #{link_id}")
    return True


def clear_backend(
    link_id: int, *, force: bool = False, hard_fault: bool = False
) -> bool:
    """Clear a link (backend)."""
    # Get the link object by link ID
    try:
        link = CircuitPythonLink.load_link_by_num(link_id)
    except FileNotFoundError as err:
        print(f"Link #{link_id} does not exist!")
        raise Exit(1) from err

    # If the link is not marked as stop, announce "--force" is needed
    if not link.stopped and not force:
        print("Can only clear links marked as inactive.")
        print(f"To force clear link #{link.link_id}, use the --force option.")
        if hard_fault:
            raise Exit(1)
        return False

    # Remove the file associated with the link
    os.remove(link.link_id_to_filename(link_id))
    print(f"Removed link #{link_id} from history")

    # Remove file from ledger, just in case
    for entry in iter_ledger_entries():
        if entry.link_id == link_id:
            remove_from_ledger(entry, expect_entry=True, use_lock=False)

    workspace.set_cws_name("")

    return True



# SPDX-FileCopyrightText: 2022 Alec Delaney
#
# SPDX-License-Identifier: MIT

"""
The sub-script for handling the workspace options for ``circlink``.

Author(s): Alec Delaney (Tekktrik)
"""

import os
import pathlib
import shutil
import zipfile
from typing import Dict, Optional

from typer import Argument, Exit, Option, Typer

from circlink import (
    CURRENT_WORKSPACE_FILE,
    LINKS_DIRECTORY,
    WORKSPACE_LIST_DIRECTORY,
)
from circlink.link import CircuitPythonLink, get_links_list

workspace_app = Typer(
    add_completion=False,
    no_args_is_help=True,
    help="Save and load workspace settings.",
)


def get_cws_name() -> str:
    """Get the current workspace name."""
    with open(CURRENT_WORKSPACE_FILE, encoding="utf-8") as cwsfile:
        name = cwsfile.read()
    return None if not name else name


def set_cws_name(name: str) -> None:
    """Set the current workspace name."""
    with open(CURRENT_WORKSPACE_FILE, mode="w", encoding="utf-8") as cwsfile:
        cwsfile.write(name)


def get_workspaces() -> Dict[str, pathlib.Path]:
    """Iterate through the workspace names and folder filepaths."""
    workspaces = [
        path
        for path in pathlib.Path(WORKSPACE_LIST_DIRECTORY).glob("*")
        if path.is_dir()
    ]
    return {path.name: path for path in workspaces}


def _remove_workspace(name: str) -> None:
    """Remove a saved workspace (if it exists)."""
    for ws_name, ws_path in get_workspaces().items():
        if name == ws_name:
            shutil.rmtree(str(ws_path.resolve()))

    set_cws_name("")


def _ensure_new_workspace(name: str, raise_error: bool = True) -> bool:
    """Save a workspace (backend)."""
    if name in get_workspaces():
        if raise_error:
            print(f"Cannot use name {name} - workspace already saved with this name.")
            raise Exit(1)
        return False
    return True


def _get_ws_path(name: str) -> str:
    """Get the for a workspace directory name."""
    return os.path.join(WORKSPACE_LIST_DIRECTORY, name)


@workspace_app.command()
def current() -> None:
    """Get the current workspace name."""
    name = get_cws_name()
    if not name:
        print("Current workspace is not named")
        raise Exit()
    print(name)


@workspace_app.command(name="list")
def workspace_list() -> None:
    """List all existing workspaces."""
    workspaces = get_workspaces()
    if workspaces:
        for workspace in get_workspaces():
            print_text = "* " + workspace if workspace == get_cws_name() else workspace
            print(print_text)
    else:
        print("No workspaces saved")


@workspace_app.command()
def delete(name: str = Argument(..., help="Name of the workspace to delete")) -> None:
    """Delete a workspace."""
    if name not in get_workspaces():
        print(f"Workspace '{name}' does not exist")
        raise Exit(1)

    _remove_workspace(name)
    print(f"Workspace '{name}' deleted")


@workspace_app.command()
def rename(
    old_name: str = Argument(..., help="Name of the workspace to rename"),
    new_name: str = Argument(..., help="New name for the workspace"),
) -> None:
    """Rename a workspace."""
    _ensure_new_workspace(new_name)

    old_path = _get_ws_path(old_name)

    parent = os.path.dirname(old_path)
    new_path = os.path.join(parent, new_name)

    os.rename(old_path, new_path)

    if get_cws_name() == old_name:
        set_cws_name(new_name)

    print(f"Workspace '{old_name}' renamed to '{new_name}'")


@workspace_app.command()
def save(
    name: str = Argument(..., help="The name of the new workspace"),
    *,
    overwrite: bool = Option(
        False,
        "--overwrite",
        "-o",
        help="Whether to overwrite an existing workspace with the same name",
    ),
) -> None:
    """Save the current link state as a workspace."""
    if not get_links_list("*"):
        print("No links are in the history, nothing to save")
        raise Exit(1)

    _ = not overwrite and _ensure_new_workspace(name)

    _remove_workspace(name)

    new_ws_folder = _get_ws_path(name)
    os.mkdir(new_ws_folder)
    links_path = pathlib.Path(LINKS_DIRECTORY)

    for link_index, link_path in enumerate(links_path.glob("*")):
        link = CircuitPythonLink.load_link_by_filepath(str(link_path))
        link._link_id = link_index + 1  # pylint: disable=protected-access
        link.process_id = 0
        link.end_flag = True
        link.confirmed = True
        link._stopped = True  # pylint: disable=protected-access
        link.save_link(save_directory=new_ws_folder)

    set_cws_name(name)

    print(f"New workspace saved as '{name}'")


@workspace_app.command()
def load(name: str = Argument(..., help="Name of the workspace to load")) -> None:
    """Load a workspace."""
    links_folder = os.path.join(LINKS_DIRECTORY)
    links_path = pathlib.Path(links_folder)
    ws_path = pathlib.Path(WORKSPACE_LIST_DIRECTORY) / name

    if list(links_path.glob("*")):
        print("Cannot load workspace with files in the history.")
        print("Please clear the history with the clear command.")
        raise Exit(1)

    if _ensure_new_workspace(name, raise_error=False):
        print("This workspace does not exist!")
        raise Exit(1)

    for link_path in ws_path.glob("*"):
        link = CircuitPythonLink.load_link_by_filepath(str(link_path))
        link.save_link()

    set_cws_name(name)

    print(f"Loaded workspace '{name}'")


@workspace_app.command()
def export(
    name: str = Argument(..., help="Name of the workspace to export"),
    path: str = Argument(
        ..., help="The folder where the workspace will be exported to"
    ),
) -> None:
    """Export a workspace."""
    export_folder = _get_ws_path(name)
    export_paths = [path.resolve() for path in pathlib.Path(export_folder).glob("*")]
    exported_path = os.path.join(path, name + ".zip")

    with zipfile.ZipFile(
        exported_path, mode="w", compression=zipfile.ZIP_DEFLATED
    ) as zip_file:
        for export_path in export_paths:
            zip_file.write(str(export_path), arcname=export_path.name)


@workspace_app.command(name="import")
def workspace_import(
    filepath: str = Argument(..., help="Filepath of the packaged workspace"),
    name: Optional[str] = Option(
        None, "--name", "-n", help="A name to give to the imported workspace"
    ),
) -> None:
    """Import a workspace."""
    package_name = os.path.basename(filepath)[:-4] if not name else name
    package_ext = os.path.splitext(filepath)[1]

    if package_ext.lower() != ".zip":
        print("Imported workspaces must be ZIP files")
        raise Exit(1)

    _ensure_new_workspace(package_name)
    new_ws_path = _get_ws_path(package_name)
    os.mkdir(new_ws_path)

    with zipfile.ZipFile(filepath) as zip_file:
        zip_file.extractall(new_ws_path)
